#include <stdio.h>
#include <stdlib.h>
#include <math.h>

struct punto {
	double x, y;
};

typedef struct punto Punto;

struct datos {
	double fx, fy, s, u0, v0;
	double theta1, theta2, theta3;
	double t1, t2, t3;
	int n;
	Punto p[16];
};

typedef struct datos Datos;

void calcula( Datos *d )
{
	int i;
	double cx, cy, w;
	double M[3][3];
	double R[3][3];
	double s1, c1, s2, c2, s3, c3;

	// Centroid:
	cx = cy = 0.0;
	for (i=0; i<d->n; i++ ) {
		cx += d->p[i].x;
		cy += d->p[i].y;
	}
	cx /= d->n;
	cy /= d->n;

	// Points centered at centroid:
	for (i=0; i<d->n; i++ ) {
		d->p[i].x -= cx;
		d->p[i].y -= cy;
	}
	d->theta1 *= M_PI/180.0;
	d->theta2 *= M_PI/180.0;
	d->theta3 *= M_PI/180.0;

	s1 = sin( d->theta1 );
	c1 = cos( d->theta1 );
	s2 = sin( d->theta2 );
	c2 = cos( d->theta2 );
	s3 = sin( d->theta3 );
	c3 = cos( d->theta3 );

	/**
          [ c1 c2 c3 - s1 s3  - c1 s3 - c2 c3 s1  c3 s2 ]
          [                                             ]
      R = [ c1 c2 s3 + c3 s1   c1 c3 - c2 s1 s3   s2 s3 ]
          [                                             ]
          [     - c1 s2             s1 s2          c2   ]
	**/

	R[0][0] = c1*c2*c3 - s1*s3;
	R[0][1] = -c1*s3 - c2*c3*s1;
	R[0][2] = c3*s2;
	
	R[1][0] = c1*c2*s3 + c3*s1;
	R[1][1] = c1*c3 - c2*s1*s3;
	R[1][2] = s2*s3;

	R[2][0] = -c1*s2;
	R[2][1] =  s1*s2;
	R[2][2] =  c2;
	/**
		M = K [ R| t] =
               [ r20 u0 + r10 s + fx r00 ]         [ r21 u0 + r11 s + fx r01 ]
               [                         ]         [                         ]
(%o8)  Col 1 = [     r20 v0 + fy r10     ] Col 2 = [     r21 v0 + fy r11     ]
               [                         ]         [                         ]
               [           r20           ]         [           r21           ]
                   [ r22 u0 + r12 s + fx r02 ]         [ t3 u0 + s t2 + fx t1 ]
                   [                         ]         [                      ]
           Col 3 = [     r22 v0 + fy r12     ] Col 4 = [    t3 v0 + fy t2     ]
                   [                         ]         [                      ]
                   [           r22           ]         [          t3          ]
	**/
	M[0][0] = R[2][0] * d->u0 + R[1][0] * d->s + d->fx * R[0][0];
	M[0][1] = R[2][1] * d->u0 + R[1][1] * d->s + d->fx * R[0][1];
	M[0][2] = d->t3 * d->u0 + d->s * d->t2  + d->fx * d->t1;

	M[1][0] = R[2][0] * d->v0 + d->fy * R[1][0];
	M[1][1] = R[2][1] * d->v0 + d->fy * R[1][1];
	M[1][2] = d->t3 * d->v0 + d->fy * d->t2;

	M[2][0] = R[2][0];
	M[2][1] = R[2][1];
	M[2][2] = d->t3;

	for( i=0; i<d->n; i++ ) {
		cx = d->p[i].x * M[0][0] + d->p[i].y * M[0][1] + M[0][2];
		cy = d->p[i].x * M[1][0] + d->p[i].y * M[1][1] + M[1][2];
		w  = d->p[i].x * M[2][0] + d->p[i].y * M[2][1] + M[2][2];

		d->p[i].x = cx / w;
		d->p[i].y = cy / w;
	}
}
	
int main( int argc, char *argv[] )
{
	int n, i;
	Datos d;

	//	octave -qf  generateImage.m $fx $fy $s $u0 $v0 $theta1 $theta2 $theta3 $t1 $t2 $t3 $inputPointSet
	if (argc < 14 ) {
		fprintf( stderr, "Args: fx fy s u0 v0 theta1 theta2 theta3 t1 t2 t3 x1 y1 ... xn yn\n" );
		return 1;
	}

	d.fx = strtod( argv[1], (char **)NULL);
	d.fy = strtod( argv[2], (char **)NULL);
	d.s = strtod( argv[3], (char **)NULL);
	d.u0 = strtod( argv[4], (char **)NULL);
	d.v0 = strtod( argv[5], (char **)NULL);

	d.theta1 = strtod( argv[6], (char **)NULL);
	d.theta2 = strtod( argv[7], (char **)NULL);
	d.theta3 = strtod( argv[8], (char **)NULL);

	d.t1 = strtod( argv[9], (char **)NULL);
	d.t2 = strtod( argv[10], (char **)NULL);
	d.t3 = strtod( argv[11], (char **)NULL);


	n = argc - 12;
	n = n/2;

	d.n = n;
	for( i=0; i<n; i++ ) {
		d.p[i].x = strtod( argv[12 + 2*i], (char **)NULL);
		d.p[i].y = strtod( argv[12 + 2*i + 1], (char **)NULL);
	}
	calcula( &d );

	// Converting image points to pixels and printing result
	for( i=0; i<d.n; i++ )
		printf( "%d %d ", (int)(d.p[i].x + 0.5), (int)(d.p[i].y + 0.5) );

	return 0;
}
